#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>

#include "memcpy.h"

static void hexdump(const char *a, size_t sz, const char *msg)
{
	int i;

	printf("%s:\n", msg);

	for (i = 0; i < sz; i++) {

		if (i && !(i & 0xf))
			printf("\n");
		else if (i && !(i & 0x7))
			printf("  ");

		printf("%.2hhx ", a[i]);
	}
	printf("\n");
}

static inline int sanity_check_buffers(char *dst, char *src, unsigned size)
{
	int i = 0;

	if (!strncmp(src, dst, size))
		return 0;

	err("src <-> dst mismatch!\n");

	while (i < size) {
		char s = src[i];
		char d = dst[i];

		if (s != d)
			printf("Mismatch at %d: %.2hhx <-> %.2hhx\n", i, s, d);
		i++;
	}
	return 1;
}

static void bench_memcpy(char *dst, char *src, unsigned size, int loops)
{
	u64 xm_cycles, mm_cycles, tmp;
	int i;

	avg_xm = avg_mm = 0;

	for (i = 0; i < loops; i++) {

		mb();
		rdtscll(tmp);
		mb();

		xm_memcpy(dst, src, size);

		mb();
		rdtscll(xm_cycles);
		mb();

		if (sanity_check_buffers(dst, src, size))
			continue;

		xm_cycles -= tmp;
		avg_xm += xm_cycles;

		dbg("XM: size: %u, cycles: %llu\n", size, xm_cycles);

		mb();
		rdtscll(tmp);
		mb();

		memcpy_64(dst, src, size);

		mb();
		rdtscll(mm_cycles);
		mb();

		mm_cycles -= tmp;
		avg_mm += mm_cycles;

		dbg("MM: size: %u, cycles: %llu\n", size, mm_cycles);
	}

	avg_xm /= loops;
	avg_mm /= loops;

	printf("%d\t%g\t\t%g\t\t%.10g\n",
		size, avg_xm, avg_mm, avg_mm / avg_xm);
}

static void do_benchmark(unsigned size, unsigned loops, int rfd)
{
	char *src, *dst;
	unsigned char unalign = 0;

	/* this makes the buffers unequally misaligned */
	unsigned char misalign = rand() % 16;
	misalign = 0;

	src = malloc(size);
	if (!src)
		perror("malloc: src");

	dst = malloc(size);
	if (!dst)
		perror("malloc: dst");

	if (read(rfd, src, size) != size)
		perror("urandom read");

	if (read(rfd, &unalign, sizeof(unalign)) != sizeof(unalign))
		perror("urandom read unalign");

	unalign %= 16;

	src += unalign;
	dst += unalign + misalign;
	size -= unalign + misalign;

	dbg("size: %d: src: %p, dst: %p, unalign: %d, misalgn: %d",
	    size, src, dst, unalign, misalign);

	bench_memcpy(dst, src, size, loops);

#if 0
	hexdump(src, size, "Source buffer");
	hexdump(dst, size, "Dest buffer");
#endif

	src -= unalign;
	dst -= unalign + misalign;

	free(src);
	free(dst);
}

int main()
{
	int rfd, i, sz, loops = 10000;

	rfd = open("/dev/urandom", O_RDONLY);
	if (rfd < 0)
		perror("urandom open");

	printf("Benchmarking with %d iterations, average results:\n", loops);
	printf("size\tXM\t\tMM\t\tspeedup\n");

	for (i = 7; i < 16; i++) {
		sz = 1 << i;

		do_benchmark(sz, loops, rfd);
		do_benchmark(sz + sz/2, loops, rfd);
		do_benchmark(sz + sz/2 + sz/5, loops, rfd);
		do_benchmark(sz + sz/2 + sz/4, loops, rfd);
		do_benchmark(sz + sz/2 + sz/3, loops, rfd);
	}
	close(rfd);

	return 0;
}

